# Sizing across corners

from definitions import *
from pyopus.design.cbd import CornerBasedDesign, generateCorners
from pyopus.evaluator.aggregate import formatParameters
# If MPI is imported an application not using MPI will behave correctly
# (i.e. only slot 0 will run the program) even when started with mpirun/mpiexec
from pyopus.parallel.mpi import MPI
from pyopus.parallel.cooperative import cOS


if __name__=='__main__':
	# Prepare statistical parameters dictionary with nominal values (0)
	nominalStat={ name: 0.0 for name, desc in statParams.items() }
	
	# Prepare operating parameters dictionary with nominal values
	nominalOp={ name: desc['init'] for name, desc in opParams.items() }
	
	# Prepare initial design parameters dictionary
	# Uncomment this if you want to use initial values from definitions.py
	# initialDesign={ name: desc['init'] for name, desc in designParams.items() }
	# Uncomment this if you want to use lower bounds for initial values of design parameters
	# initialDesign={ name: desc['lo'] for name, desc in designParams.items() }
	# Uncomment this if you want to use the result of nominal design as initial point
	initialDesign={
		'c_out':    2.0125484638530715e-12, 
		'diff_l':   9.0173130168525476e-07, 
		'diff_w':   1.3748137933373763e-05, 
		'load_l':   3.1959280853504449e-06, 
		'load_w':   8.9130317977231035e-05, 
		'mirr_l':   1.9427037931893832e-06, 
		'mirr_ld':  2.4749186739319866e-06, 
		'mirr_w':   7.9105009110997434e-05, 
		'mirr_wd':  2.0887405636991897e-05, 
		'mirr_wo':  3.8316438575060692e-05, 
		'out_l':    2.7397889227076129e-06, 
		'out_w':    1.3868429320545923e-05, 
		'r_out':    1.7894053483506941e+04, 
	}
	
	# Prepare corners, cartesian product of 
	#   MOS model:    wp, ws, wo, and wz (named wp, ws, wo, and wz)
	#   voltages:     1.7 and 2.0        (named vl and vh)
	#   temperatures: 0 and 100          (named tl and th)
	corners, names = generateCorners(
		specs=[
			(
				'model', 'mos', 
				['wp', 'ws', 'wo', 'wz'], 
				['wp', 'ws', 'wo', 'wz']
			), 
			(
				'param', 'vdd', 
				[opParams['vdd']['lo'], opParams['vdd']['hi']], 
				['vl', 'vh'], 
			), 
			(
				'param', 'temperature', 
				[opParams['temperature']['lo'], opParams['temperature']['hi']], 
				['tl', 'th'], 
			), 
		], 
		heads=[ 'opus' ], 
	)
	
	# Add nominal corner (model tm, 1.8V, 25degC), defined for head opus
	nominalCorner={
		('nom',('opus',)): {
			'modules': [ 'tm' ], 
			'params': {
				'vdd': opParams['vdd']['init'], 
				'temperature': opParams['temperature']['init']
			}, 
		}
	}
	corners.update(nominalCorner)
	
	# Now we have 4x2x2+1=17 corners, print their names
	print("Corner names")
	print(sorted(corners.keys()))
	
	# Area should be evauated only in corner 'nom'
	measures['area']['corners']=[ 'nom' ]
	
	# Prepare parallel environment
	cOS.setVM(MPI(mirrorMap={'*':'.'}, persistentStorage=False))
	
	# Design it, nominal statistical parameters are treated as fixed parameters
	cbd=CornerBasedDesign(
		designParams, heads, analyses, measures, corners, 
		fixedParams=[nominalStat], variables=variables, norms={ 'area': 100e-12 }, 
		# tradeoff=1e-6, 
		# stepTol=1e-4, 
		initial=initialDesign, 
		method='global', 
		incrementalCorners=True, 
		evaluatorOptions={'debug': 0}, 
		debug=1
	)
	
	# Run optimization
	atDesign, aggregator, analysisCount = cbd()
	
	# Print design parameters and performance measures
	print(formatParameters(atDesign))
	print(aggregator.formatResults())
	
	# Finalize cOS parallel environment
	cOS.finalize()
	
# Result
#           c_out:   6.4025057781114802e-12
#          diff_l:   3.4247816773735976e-07
#          diff_w:   1.6738092946223754e-05
#          load_l:   2.3668072815514710e-06
#          load_w:   5.3240803355989389e-05
#          mirr_l:   1.0157035640188838e-06
#         mirr_ld:   1.6182520447950098e-06
#          mirr_w:   5.8107861163445226e-05
#         mirr_wd:   6.1242813846216839e-05
#         mirr_wo:   6.8277846319352392e-05
#           out_l:   4.1088626829554061e-07
#           out_w:   8.1812469568552345e-05
#           r_out:   5.1647907484619873e+02
